/*!
    Actions is an enumeration of any actions that is possible to do 
    on the controls. 
*/
/*
    Copyright (C) 2016  Gabriel Dubé

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


use std::hash::Hash;
use constants::{Error, CheckState, WindowDisplay, MessageButtons, MessageIcons, ControlType, MessageChoice};

#[derive(PartialEq)]
pub struct ActMessageParams {
    pub title: String,
    pub content: String,
    pub buttons: MessageButtons,
    pub icons: MessageIcons
}

/**
    Possible message to send to an Ui
*/
#[derive(PartialEq)]
pub enum Action<ID: Eq+Clone+Hash> {
    None,
    Reset,
    GetControlType,
    Undo,
    Close,
        
    GetParent,
    SetParent(Option<Box<ID>>),
    GetChildren,
    GetDescendants,
    
    GetPosition,
    SetPosition(i32, i32),
    
    GetSize,
    SetSize(u32, u32),
    
    GetText,
    SetText(Box<String>),
    
    GetCheckState,
    SetCheckState(CheckState),
    
    GetEnabled,
    SetEnabled(bool),
    
    GetVisibility,
    SetVisibility(bool),
    
    GetWindowDisplay,
    SetWindowDisplay(WindowDisplay),
    
    GetTextLimit,
    SetTextLimit(u32),
    
    GetSelectedBounds,
    SetSelectedBounds((u32, u32)),
    
    GetReadonly,
    SetReadonly(bool),
    
    GetSelectedIndex,
    SetSelectedIndex(Option<u32>),

    GetPlaceholder,
    SetPlaceholder(Option<Box<String>>),

    GetDropdownVisibility,
    SetDropdownVisibility(bool),
    
    AddString(Box<String>),
    RemoveString(Box<String>),
    FindString(Box<String>),
    InsertString(Box<(u32, String)>),
    GetIndexedString(u32),
    
    GetStringCollection,
    SetStringCollection(Box<Vec<String>>),

    CountItems,
    RemoveIndexedItem(u32),

    Message(Box<ActMessageParams>)
}

/**
    Possible values returned by message sent to an Ui
*/
#[derive(PartialEq)]
pub enum ActionReturn<ID: Eq+Clone+Hash> {
    None,
    Parent(Box<ID>),
    Children(Box<Vec<ID>>),
    Position(i32, i32),
    Size(u32, u32),
    Text(Box<String>),
    Error(Error),
    CheckState(CheckState),
    Enabled(bool),
    Visibility(bool),
    WindowDisplay(WindowDisplay),
    TextLimit(u32),
    SelectBounds((u32, u32)),
    Readonly(bool),
    ItemIndex(u32),
    ItemCount(u32),
    StringCollection(Box<Vec<String>>),
    ControlType(ControlType),
    MessageChoice(MessageChoice),
    NotSupported
}


pub mod helper {
    use actions::{Action, ActMessageParams};
    use constants::{MessageButtons, MessageIcons};
    use std::hash::Hash;

    /**
        Action helper for the Message action.
    */
    #[inline(always)]
    pub fn message<ID: Eq+Clone+Hash, S1: Into<String>, S2: Into<String>>(title: S1, content: S2, buttons: MessageButtons, icons: MessageIcons) -> Action<ID> {
        Action::Message(Box::new(ActMessageParams{
            title: title.into(),
            content: content.into(),
            buttons: buttons,
            icons: icons
        }))
    }

    /**
        Action helper for the SetText action.
    */
    #[inline(always)]
    pub fn set_text<ID: Eq+Clone+Hash, S: Into<String>>(text: S) -> Action<ID> {
        Action::SetText(Box::new(text.into()))
    }

    /**
        Action helper for the SetParent action.
    */
    #[inline(always)]
    pub fn set_parent<ID: Eq+Clone+Hash>(p: ID) -> Action<ID> {
        Action::SetParent(Some(Box::new(p)))
    }

    /**
        Action helper for the SetParent action.
    */
    #[inline(always)]
    pub fn remove_parent<ID: Eq+Clone+Hash>() -> Action<ID> {
        Action::SetParent(None)
    }

    /**
        Action helper for AddString
    */
    #[inline(always)]
    pub fn add_string<ID: Eq+Clone+Hash, S: Into<String>>(s: S) -> Action<ID> {
        Action::AddString(Box::new(s.into()))
    }

    /**
        Action helper for FindString
    */
    #[inline(always)]
    pub fn find_string<ID: Eq+Clone+Hash, S: Into<String>>(s: S) -> Action<ID> {
        Action::FindString(Box::new(s.into()))
    }

    /**
        Action helper for RemoveString
    */
    #[inline(always)]
    pub fn remove_string<ID: Eq+Clone+Hash, S: Into<String>>(s: S) -> Action<ID> {
        Action::RemoveString(Box::new(s.into()))
    }

    /**
        Action helper for SetSelectedIndex
    */
    #[inline(always)]
    pub fn set_selected_index<ID: Eq+Clone+Hash>(i: u32) -> Action<ID> {
        Action::SetSelectedIndex(Some(i))
    }

    /**
        Action helper for SetSelectedIndex
    */
    #[inline(always)]
    pub fn remove_selected<ID: Eq+Clone+Hash>() -> Action<ID> {
        Action::SetSelectedIndex(None)
    }

    /**
        Action helper to set a placeholder on a control
    */
    #[inline(always)]
    pub fn set_placeholder<ID: Eq+Clone+Hash, S: Into<String>>(s: S) -> Action<ID> {
        Action::SetPlaceholder(Some(Box::new(s.into())))
    }

    /**
        Action helper to set a placeholder on a control
    */
    #[inline(always)]
    pub fn remove_placeholder<ID: Eq+Clone+Hash>() -> Action<ID> {
        Action::SetPlaceholder(None)
    }

    /**
        Action helper for insert string
    */
    #[inline(always)]
    pub fn insert_string<ID: Eq+Clone+Hash, S: Into<String>>(i: u32, s: S) -> Action<ID> {
        Action::InsertString(Box::new((i, s.into())))
    }

}